from enum import Enum, unique

from data.registry_model.datatype import datatype, requiresinput
from image.docker.schema1 import DockerSchema1Manifest


class RepositoryReference(datatype('Repository', [])):
  """ RepositoryReference is a reference to a repository, passed to registry interface methods. """
  @classmethod
  def for_repo_obj(cls, repo_obj):
    if repo_obj is None:
      return None

    return RepositoryReference(db_id=repo_obj.id)


class Label(datatype('Label', ['key', 'value', 'uuid', 'source_type_name', 'media_type_name'])):
  """ Label represents a label on a manifest. """
  @classmethod
  def for_label(cls, label):
    if label is None:
      return None

    return Label(db_id=label.id, key=label.key, value=label.value,
                 uuid=label.uuid, media_type_name=label.media_type.name,
                 source_type_name=label.source_type.name)


class Tag(datatype('Tag', ['name', 'reversion', 'manifest_digest', 'lifetime_start_ts',
                           'lifetime_end_ts'])):
  """ Tag represents a tag in a repository, which points to a manifest or image. """
  @classmethod
  def for_repository_tag(cls, repository_tag, manifest_digest=None, legacy_image=None):
    if repository_tag is None:
      return None

    return Tag(db_id=repository_tag.id,
               name=repository_tag.name,
               reversion=repository_tag.reversion,
               lifetime_start_ts=repository_tag.lifetime_start_ts,
               lifetime_end_ts=repository_tag.lifetime_end_ts,
               manifest_digest=manifest_digest,
               inputs=dict(legacy_image=legacy_image))

  @property
  @requiresinput('legacy_image')
  def legacy_image(self, legacy_image):
    """ Returns the legacy Docker V1-style image for this tag. Note that this
        will be None for tags whose manifests point to other manifests instead of images.
    """
    return legacy_image


class Manifest(datatype('Manifest', ['digest', 'manifest_bytes'])):
  """ Manifest represents a manifest in a repository. """
  @classmethod
  def for_tag_manifest(cls, tag_manifest, legacy_image=None):
    if tag_manifest is None:
      return None

    return Manifest(db_id=tag_manifest.id, digest=tag_manifest.digest,
                    manifest_bytes=tag_manifest.json_data,
                    inputs=dict(legacy_image=legacy_image))

  @property
  @requiresinput('legacy_image')
  def legacy_image(self, legacy_image):
    """ Returns the legacy Docker V1-style image for this manifest. Note that this
        will be None for manifests that point to other manifests instead of images.
    """
    return legacy_image

  def get_parsed_manifest(self, validate=True):
    """ Returns the parsed manifest for this manifest. """
    return DockerSchema1Manifest(self.manifest_bytes, validate=validate)


class LegacyImage(datatype('LegacyImage', ['docker_image_id', 'created', 'comment', 'command',
                                           'image_size', 'aggregate_size', 'uploading'])):
  """ LegacyImage represents a Docker V1-style image found in a repository. """
  @classmethod
  def for_image(cls, image, images_map=None, tags_map=None):
    if image is None:
      return None

    return LegacyImage(db_id=image.id,
                       inputs=dict(images_map=images_map, tags_map=tags_map,
                                   ancestor_id_list=image.ancestor_id_list()),
                       docker_image_id=image.docker_image_id,
                       created=image.created,
                       comment=image.comment,
                       command=image.command,
                       image_size=image.storage.image_size,
                       aggregate_size=image.aggregate_size,
                       uploading=image.storage.uploading)

  @property
  @requiresinput('images_map')
  @requiresinput('ancestor_id_list')
  def parents(self, images_map, ancestor_id_list):
    """ Returns the parent images for this image. Raises an exception if the parents have
        not been loaded before this property is invoked.
    """
    return [LegacyImage.for_image(images_map[ancestor_id], images_map=images_map)
            for ancestor_id in reversed(ancestor_id_list)
            if images_map.get(ancestor_id)]

  @property
  @requiresinput('tags_map')
  def tags(self, tags_map):
    """ Returns the tags pointing to this image. Raises an exception if the tags have
        not been loaded before this property is invoked.
    """
    tags = tags_map.get(self._db_id)
    if not tags:
      return []

    return [Tag.for_repository_tag(tag) for tag in tags]


@unique
class SecurityScanStatus(Enum):
  """ Security scan status enum """
  SCANNED = 'scanned'
  FAILED = 'failed'
  QUEUED = 'queued'
